package ij.plugin; 

import ij.*; 
import ij.gui.GenericDialog; 
import ij.process.*; 

import java.lang.*; 
import java.awt.*; 
import java.awt.event.*; 


/** This plugin performs a z-projection of the input stack. Type of
    output image is same as type of input image. Both maximum and
    average intensity projections are supported. 

    @author Patrick Kelly <phkelly@ucsd.edu> */


public class ZProjector implements PlugIn {
    public static final int AVG_METHOD = 0; 
    public static final int MAX_METHOD = 1;
    public static final int SUM_METHOD = 2;
    private String[] items = {"Average Intensity", "Max Intensity", "Sum Slices"}; 
    private static int method = AVG_METHOD;

    private static final int BYTE_TYPE  = 0; 
    private static final int SHORT_TYPE = 1; 
    private static final int FLOAT_TYPE = 2; 

    /** Image to hold z-projection. */
    private ImagePlus projImage = null; 

    /** Image stack to project. */
    private ImagePlus imp = null; 

    /** Projection starts from this slice. */
    private int startSlice = 1;
    /** Projection ends at this slice. */
    private int stopSlice = 1; 

    // --------------------------------------------------
    public ZProjector() {
    }

    // --------------------------------------------------
    /** Construction of ZProjector with image to be projected. */
    public ZProjector(ImagePlus imp) {
	setImage(imp); 
    }

    // --------------------------------------------------
    /** Explicitly set image to be projected. This is useful if
	ZProjection_ object is to be used not as a plugin but as a
	stand alone processing object.  */
    public void setImage(ImagePlus imp) {
    	this.imp = imp; 

	startSlice = 1; 
	stopSlice = imp.getStackSize(); 
    }

    // --------------------------------------------------
    public void setStartSlice(int slice) {
	if(imp==null || slice < 1 || slice > imp.getStackSize())
	    return; 

	startSlice = slice; 
    }

    // --------------------------------------------------
    public void setStopSlice(int slice) {
	if(imp==null || slice < 1 || slice > imp.getStackSize())
	    return; 

	stopSlice = slice; 
    }

    // --------------------------------------------------
    /** Retrieve results of most recent projection operation.
     */
    public ImagePlus getProjection() {
	return projImage; 
    }


    // --------------------------------------------------
    public void run(String arg) {
	imp = WindowManager.getCurrentImage();
	
	if(imp==null) {
	    IJ.noImage(); 
	    return; 
	}

	//  Make sure input image is a stack.
	if(imp.getStackSize()==1) {
	    IJ.error("ZProjection: must call this plugin on image stack."); 
	    return; 
	}
	
	// Set default bounds.
	startSlice = 1; 
	stopSlice  = imp.getStackSize(); 

	// Build control dialog.
	GenericDialog gd = buildControlDialog(startSlice,stopSlice); 
	gd.showDialog(); 

	if(gd.wasCanceled()) return; 

	if(!imp.lock()) return;   // exit if in use
	long tstart = System.currentTimeMillis(); 
	doProjection(gd); 

	if(arg.equals("")) {
		long tstop = System.currentTimeMillis(); 
	    projImage.show("ZProjector: " +IJ.d2s((tstop-tstart)/1000.0,2)+" seconds");
	}

	imp.unlock(); 
	IJ.register(ZProjector.class);

	return; 
    }

    // --------------------------------------------------
    /** Reads values from dialog and performs projection.
	@param gd read values from this generic dialog
    */
    private void doProjection(GenericDialog gd) {
	// Update starting and stopping slice values.
	setStartSlice((int)gd.getNextNumber()); 
	setStopSlice((int)gd.getNextNumber()); 

	// Get projection method.
	method = gd.getNextChoiceIndex(); 

	computeProjection(method); 
    }

    // --------------------------------------------------
    /** Builds dialog to query users for projection parameters.
	@param start starting slice to display
	@param stop last slice */
    protected GenericDialog buildControlDialog(int start, int stop) {
	GenericDialog gd = new GenericDialog("ZProjection",IJ.getInstance()); 
	gd.addNumericField("Start slice:",startSlice,0/*digits*/); 
	gd.addNumericField("Stop slice:",stopSlice,0/*digits*/); 

	// Different kinds of projections.
	gd.addChoice("Projection Type",items,items[method]); 

	return gd; 
    }

    // --------------------------------------------------
    /** Performs actual projection using specified method. 
 	@param method specifies projection method to use
	@param show flag indicating whether projected image should be
	displayed */
    public void computeProjection(int method) {
	if(imp==null) return; 

	projImage = null; 

	if(method==AVG_METHOD||method==SUM_METHOD) {
	    // Set up an average intensity projection.
	    AverageIntensity ai = new AverageIntensity(stopSlice-startSlice+1); 
	    projImage = projectStack(ai); 
	}
	else if(method==MAX_METHOD) {
	    // Set up a max intensity projection.
	    MaxIntensity mi = new MaxIntensity(); 
	    projImage = projectStack(mi); 
	}
	else
	    IJ.error("ZProjection - unknown method."); 

	if(projImage==null) {
	    IJ.error("ZProjection - error computing projection."); 
	    return; 
	}
    }

    // --------------------------------------------------
    private ImagePlus projectStack(RayFunction rayFunc) {
	if(IJ.debugMode==true) {
	    IJ.write("\nProjecting stack from: "+startSlice
		     +" to: "+stopSlice); 
	}

	// Width and height of input image.
	int width  = imp.getWidth();
	int height = imp.getHeight(); 

	// Create new float processor using projected pixels.
	FloatProcessor fp = new FloatProcessor(width,height); 
	float[] pixels = (float[])fp.getPixels(); 
	ImageStack stack = imp.getStack(); 

	// Determine type of input image. Explicit determination of
	// processor type is required for subsequent pixel
	// manipulation.  This approach is more efficient than the
	// more general use of ImageProcessor's getPixelValue and
	// putPixel methods.
	int ptype; 
	if(stack.getProcessor() instanceof ByteProcessor)       ptype = BYTE_TYPE; 
	else if(stack.getProcessor() instanceof ShortProcessor) ptype = SHORT_TYPE; 
	else if(stack.getProcessor() instanceof FloatProcessor) ptype = FLOAT_TYPE; 
	else {
	    IJ.error("ZProjection_: Unknown processor type."); 
	    return null; 
	}

	// Directly traversing pixel array is more efficient than
	// using ImageProcessor pixel acces routines.
	Object[] pixelArray = stack.getImageArray(); 

	if (rayFunc instanceof MaxIntensity)
		for (int i=0; i<pixels.length; i++)
			pixels[i] = -Float.MAX_VALUE;
		
	// Do the projection.
	for(int n=startSlice; n<=stopSlice;++n) {
	    IJ.showStatus("ZProjection - projecting slice: "+n); 
	    projectSlice(pixels,pixelArray[n-1],rayFunc,ptype); 
	}

	// Finish up projection.
	ImagePlus rtnImg;
	if (method==SUM_METHOD) {
		if (imp.getProcessor().isUnsigned())
			fp.add(32768*stack.getSize());
		fp.setMinAndMax(0.0,0.0);
		rtnImg = new ImagePlus("Sum",fp); 
	} else {
		rayFunc.postProcess(fp); 
		rtnImg = makeOutputImage(imp,fp,ptype);
	} 

	return rtnImg; 
    }

    // --------------------------------------------------
    /** Generate output image whose type is same as input image. */
    private ImagePlus makeOutputImage(ImagePlus imp, FloatProcessor fp, int ptype) {
		int width = imp.getWidth(); 
		int height = imp.getHeight(); 
		float[] pixels = (float[])fp.getPixels(); 
		ImageProcessor oip=null; 

		// Create output image consistent w/ type of input image.
		int size = pixels.length;
		switch (ptype) {
			case BYTE_TYPE:
				oip = imp.getProcessor().createProcessor(width,height);
				byte[] pixels8 = (byte[])oip.getPixels(); 
				for(int i=0; i<size; i++)
					pixels8[i] = (byte)pixels[i];
				break;
			case SHORT_TYPE:
				oip = imp.getProcessor().createProcessor(width,height);
				short[] pixels16 = (short[])oip.getPixels(); 
				for(int i=0; i<size; i++)
					pixels16[i] = (short)pixels[i];
				break;
			case FLOAT_TYPE:
				oip = new FloatProcessor(width, height, pixels, null);
				break;
		}
	
	// Adjust for display.
	if(!(oip instanceof ByteProcessor)) {
	    // Calling this on non-ByteProcessors ensures image
	    // processor is set up to correctly display image.  Byte
	    // processors do not support the same semantics so we
	    // don't envoke if type is ByteProcessor.
	    oip.setMinAndMax(0,0); 
	}

	// Create new image plus object. Don't use
	// ImagePlus.createImagePlus here because there may be
	// attributes of input image that are not appropriate for
	// projection.
	return new ImagePlus("zprojection",oip); 
    }

    // --------------------------------------------------
    /** Handles mechanics of projection by selecting appropriate pixel
	array type. We do this rather than using more general
	ImageProcessor getPixelValue() and putPixel() methods because
	direct manipulation of pixel arrays is much more efficient.  */
    private void projectSlice(float[] pixels, 
			      Object pixelArray,
			      RayFunction rayFunc,
			      int ptype)
    {
	switch(ptype) {
	case BYTE_TYPE:
	    rayFunc.projectSlice(pixels,(byte[])pixelArray); 
	    break; 
	case SHORT_TYPE:
	    rayFunc.projectSlice(pixels,(short[])pixelArray); 
	    break; 
	case FLOAT_TYPE:
	    rayFunc.projectSlice(pixels,(float[])pixelArray); 
	    break; 
	}
    }

    // ==================================================
    //
    /** Abstract class that specifies structure of ray
	function. Preprocessing should be done in derived class
	constructors. Make this an abstract class rather than
	interface as abstract class seems to give slightly better
	performance.  */
    abstract class RayFunction {
	/** Perform any necessary post processing operations, e.g.
	    averging values. */
	public void postProcess(FloatProcessor fp){}

	/** Do actual slice projection for specific data
	    types. Projection data is assumed to be accumulated in
	    fpixels array. */
	public void projectSlice(float[] fpixels,byte[] pixels){ }
	public void projectSlice(float[] fpixels,short[] pixels){}
	public void projectSlice(float[] fpixels,float[] pixels){}

    } // end RayFunction


    // ==================================================
    /** Compute average intensity projection. */
    class AverageIntensity extends RayFunction {
	private int num; 

	/** Constructor requires number of slices to be
	    projected. This is used to determine average at each
	    pixel. */
	public AverageIntensity(int num) {
	    this.num = num;
	}

	// --------------------------------------------------
	public void projectSlice(float[] fpixels,byte[] pixels) {
	    int number = fpixels.length-1; 
	    for(int n=number;n>=0;--n)
			fpixels[n] += (pixels[n]&0xff); 
	}

	// --------------------------------------------------
	public void projectSlice(float[] fpixels,short[] pixels) {
	    int number = fpixels.length-1;
	    short v; 
	    for(int n=number;n>=0;--n) {
			fpixels[n] += (pixels[n]);
		} 
	}


	// --------------------------------------------------
	public void projectSlice(float[] fpixels,float[] pixels) {
	    int number = fpixels.length-1; 
	    for(int n=number;n>=0;--n)
			fpixels[n] += pixels[n]; 
	}


	// --------------------------------------------------
	// Average values.
	public void postProcess(FloatProcessor fp) {
	    float[] pixels = (float[])fp.getPixels(); 

	    int number = pixels.length-1; 
	    for(int n=number; n>=0;--n) {
			pixels[n] /= (float)num; 
	    }
	}

    } // end AverageIntensity


    // ==================================================
    /** Compute max intensity projection. */
    class MaxIntensity extends RayFunction {
	/** Simple constructor since no preprocessing is necessary. */
	public MaxIntensity() {}

	// --------------------------------------------------
	public void projectSlice(float[] fpixels, byte[] pixels) {
	    int number = fpixels.length-1; 
	    for(int n=number;n>=0;--n) {
			if((pixels[n]&0xff)>fpixels[n])
		    	fpixels[n] = (pixels[n]&0xff); 
	    }
	}

	// --------------------------------------------------
	public void projectSlice(float[] fpixels, short[] pixels) {
	    int number = fpixels.length-1;
	    for(int n=number;n>=0;--n) {
			double d = fpixels[n];
			if(pixels[n]>fpixels[n])
		    	fpixels[n] = pixels[n];
	    }
	}

	// --------------------------------------------------
	public void projectSlice(float[] fpixels, float[] pixels) {
	    int number = fpixels.length-1; 
	    for(int n=number;n>=0;--n) {
			if(pixels[n]>fpixels[n])
		    	fpixels[n] = pixels[n]; 
	    }
	}

    } // end MaxIntensity

}  // end ZProjection


